@@ -4,5 +4,6 @@ from toolformer_pytorch.toolformer_pytorch import (
     Toolformer,
     filter_tokens_with_api_response,
     sample,
-    sample_with_api_call
+    sample_with_api_call,
+    invoke_tools
 )
